/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __APP_PIKA_MEMSIZE_H__
#define __APP_PIKA_MEMSIZE_H__


gint64   pika_g_type_instance_get_memsize      (GTypeInstance   *instance);
gint64   pika_g_object_get_memsize             (GObject         *object);

gint64   pika_g_hash_table_get_memsize         (GHashTable      *hash,
                                                gint64           data_size);
gint64   pika_g_hash_table_get_memsize_foreach (GHashTable      *hash,
                                                PikaMemsizeFunc  func,
                                                gint64          *gui_size);

gint64   pika_g_slist_get_memsize              (GSList          *slist,
                                                gint64           data_size);
gint64   pika_g_slist_get_memsize_foreach      (GSList          *slist,
                                                PikaMemsizeFunc  func,
                                                gint64          *gui_size);

gint64   pika_g_list_get_memsize               (GList           *list,
                                                gint64           data_size);
gint64   pika_g_list_get_memsize_foreach       (GList           *list,
                                                PikaMemsizeFunc  func,
                                                gint64          *gui_size);

gint64   pika_g_queue_get_memsize              (GQueue          *queue,
                                                gint64           data_size);
gint64   pika_g_queue_get_memsize_foreach      (GQueue          *queue,
                                                PikaMemsizeFunc  func,
                                                gint64          *gui_size);

gint64   pika_g_value_get_memsize              (GValue          *value);
gint64   pika_g_param_spec_get_memsize         (GParamSpec      *pspec);

gint64   pika_gegl_buffer_get_memsize          (GeglBuffer      *buffer);
gint64   pika_gegl_pyramid_get_memsize         (GeglBuffer      *buffer);

gint64   pika_string_get_memsize               (const gchar     *string);
gint64   pika_parasite_get_memsize             (PikaParasite    *parasite,
                                                gint64          *gui_size);


#endif /* __APP_PIKA_MEMSIZE_H__ */
